@@ -4,19 +4,18 @@ from typing import Callable
 
 from jsonrpcserver import dispatch
 from jsonrpcserver import methods
-
-from jupyter_ascending.logger import J_LOGGER
+from loguru import logger
 
 
 def _wrap_request(f: Callable, start_msg: str, close_msg: str):
     @wraps(f)
-    @J_LOGGER.catch
+    @logger.catch
     def wrapper(*args, **kwargs):
-        J_LOGGER.debug("{}: {}", start_msg, f.__name__)
+        logger.debug("{}: {}", start_msg, f.__name__)
 
         result = f(*args, **kwargs)
 
-        J_LOGGER.debug("{}: {}", close_msg, result)
+        logger.debug("{}: {}", close_msg, result)
 
         return result
 
@@ -27,7 +26,8 @@ class ServerMethods:
     """
     Wrapper to make some things a bit nicer around jsonrpcserver.methods.Methods
 
-    Adds auto logging and error catching so that you don't have to remember to do that.
+    Basically our own version of jsonrpcserver.methods.Methods, wrapping each method with auto
+    logging and error catching so that you don't have to remember to do that.
     """
 
     def __init__(self, start_msg: str, close_msg: str):
@@ -42,15 +42,24 @@ class ServerMethods:
 
 
 def generate_request_handler(name: str, methods: ServerMethods) -> BaseHTTPRequestHandler:
-    @J_LOGGER.catch
+    """Build a handler to respond to HTTP POST requests containing JSON-RPC messages.
+
+    Will call jsonrpcserver.dispatch to dispatch the request to the appropriate handler.
+
+    TODO: why this weird construction vs a simple subclass?
+        - to be able to specify a custom class name, i think. but why do we need that?
+    """
+
+    @logger.catch
     def do_POST(self):
         # Process request
         request = self.rfile.read(int(self.headers["Content-Length"])).decode()
-        J_LOGGER.info("{} processing request:\n\t\t{}", name, request)
+        logger.info("{} processing request:\n\t\t{}", name, request)
 
+        # Dispatch the RPC request to the right function and get the function's response.
         response = dispatch(request, methods=methods)
 
-        J_LOGGER.info("Got Response:\n\t\t{}", response)
+        logger.info("Got Response:\n\t\t{}", response)
 
         # Return response
         self.send_response(response.http_status)
@@ -59,7 +68,7 @@ def generate_request_handler(name: str, methods: ServerMethods) -> BaseHTTPReque
         self.wfile.write(str(response).encode())
 
     def log_message(self, format, *args):
-        J_LOGGER.debug(args)
+        logger.debug(args)
 
     return type(
         f"{name}RequestHandler",
