@@ -1,4 +1,10 @@
+"""
+This file contains code for the JSON-RPC server we run alongside each .sync.ipynb notebook.
+
+It receives messages from `jupyter_server.py` and takes the appropriate action in the notebook.
+"""
 import threading
+import time
 from http.server import HTTPServer
 from inspect import signature
 from pathlib import Path
@@ -13,16 +19,18 @@ import jupytext
 from IPython.display import display
 from ipykernel.comm import Comm
 from jsonrpcclient import request
+from loguru import logger
 
 from jupyter_ascending._environment import EXECUTE_HOST_URL
 from jupyter_ascending.handlers import ServerMethods
 from jupyter_ascending.handlers import generate_request_handler
+from jupyter_ascending.handlers.server_extension import register_notebook_server
 from jupyter_ascending.json_requests import ExecuteAllRequest
 from jupyter_ascending.json_requests import ExecuteRequest
 from jupyter_ascending.json_requests import FocusCellRequest
 from jupyter_ascending.json_requests import GetStatusRequest
+from jupyter_ascending.json_requests import RestartRequest
 from jupyter_ascending.json_requests import SyncRequest
-from jupyter_ascending.logger import J_LOGGER
 from jupyter_ascending.notebook.data_types import JupyterCell
 from jupyter_ascending.notebook.data_types import NotebookContents
 from jupyter_ascending.notebook.merge import OpCodeAction
@@ -32,15 +40,15 @@ from jupyter_ascending.utils import find_free_port
 
 COMM_NAME = "AUTO_SYNC::notebook"
 
-_JupyterComm = None
-
 notebook_server_methods = ServerMethods("JupyterNotebook Start", "JupyterNotebook Close")
 
+merge_complete = False
 
-@J_LOGGER.catch
-def start_notebook_server_in_thread(
-    notebook_name: str, server, status_widget=None
-):
+lock = threading.Lock()
+
+
+@logger.catch
+def start_notebook_server_in_thread(notebook_name: str, status_widget=None):
     """
     Args:
         notebook_name: The name of the notebook you want to be syncing in this process.
@@ -60,26 +68,23 @@ def start_notebook_server_in_thread(
     notebook_executor_thread = threading.Thread(target=notebook_executor.serve_forever, args=tuple())
     notebook_executor_thread.start()
 
-    J_LOGGER.info("IPYTHON: Registering notebook {}", notebook_path)
+    logger.info("IPYTHON: Registering notebook {}", notebook_path)
     request(
         EXECUTE_HOST_URL,
-        server.register_notebook_server.__name__,
+        register_notebook_server.__name__,
         # Params
         notebook_path=str(notebook_path),
         port_number=notebook_server_port,
     )
-    J_LOGGER.info("==> Success")
-
-    make_comm()
+    logger.info("==> Success")
 
     return status_widget
 
 
-
 def dispatch_json_request(f):
     """
+    A kinda weird decorator attempting to remove some boilerplate in the following funcs.
     Automatically dispatch a json request based on the request_type.
-    This means we don't have to repeat ourselves.
 
     Adds it to notebook_server_methods
     """
@@ -98,9 +103,10 @@ def dispatch_json_request(f):
 
 @dispatch_json_request
 def handle_execute_request(request_type: Type[ExecuteRequest], data: dict) -> str:
+    """JSON-RPC request handler for 'execute cell'"""
     request = request_type(**data)
 
-    comm = get_comm()
+    comm = make_comm()
     execute_cell_contents(comm, request.cell_index)
 
     return f"Executing cell `{request.cell_index}`"
@@ -108,10 +114,11 @@ def handle_execute_request(request_type: Type[ExecuteRequest], data: dict) -> st
 
 @dispatch_json_request
 def handle_execute_all_request(request_type: Type[ExecuteAllRequest], data: dict) -> str:
+    """JSON-RPC request handler for 'execute all cells'"""
     request = request_type(**data)
 
-    # TODO: Remind mysefyl why I don't need to say the filename here...
-    comm = get_comm()
+    # TODO: Remind myself why I don't need to say the filename here...
+    comm = make_comm()
     execute_all_cells(comm)
 
     return f"Executing all cells in {request.file_name}"
@@ -119,20 +126,26 @@ def handle_execute_all_request(request_type: Type[ExecuteAllRequest], data: dict
 
 @dispatch_json_request
 def handle_sync_request(request_type: Type[SyncRequest], data: dict) -> str:
+    """JSON-RPC request handler for 'sync'"""
+    global merge_complete
+    merge_complete = False
     request = request_type(**data)
 
-    comm = get_comm()
+    # We lock here because updating the notebook isn't threadsafe.
+    # If we got two sync requests simultaneously without a lock,
+    # bad things might happen (eg duplicated inserts/deletes).
+    with lock:
+        comm = make_comm()
 
-    result = jupytext.reads(request.contents, fmt="py:percent")
-    # import ipdb
-    # ipdb.set_trace()
-    update_cell_contents(comm, result)
+        result = jupytext.reads(request.contents, fmt="py:percent")
+        update_cell_contents(comm, result)
 
     return "Syncing all cells"
 
 
 @dispatch_json_request
 def handle_focus_cell_request(request_type: Type[FocusCellRequest], data: dict) -> str:
+    """JSON-RPC request handler for 'focus cell'"""
     request = request_type(**data)
 
     print(request)
@@ -141,23 +154,39 @@ def handle_focus_cell_request(request_type: Type[FocusCellRequest], data: dict)
 
 @dispatch_json_request
 def handle_get_status_request(request_type: Type[GetStatusRequest], data: dict) -> str:
-    J_LOGGER.info("Attempting get_status")
+    """JSON-RPC request handler for 'get status'"""
+    logger.info("Attempting get_status")
 
-    comm = get_comm()
+    comm = make_comm()
     comm.send({"command": "get_status"})
 
-    J_LOGGER.info("Sent get_status")
+    logger.info("Sent get_status")
 
     return f"Updating status"
 
 
+@dispatch_json_request
+def handle_restart_request(request_type: Type[RestartRequest], data: dict) -> str:
+    """JSON-RPC request handler for 'restart'"""
+    request = request_type(**data)
+
+    comm = make_comm()
+    comm.send({"command": "restart_kernel"})
+    logger.info("Sent restart_kernel")
+
+    return f"Restarting kernel in {request.file_name}"
+
+
 NotebookKernelRequestHandler = generate_request_handler("NotebookKernel", notebook_server_methods)
 
 
-def make_comm() -> None:
-    global _JupyterComm
+def make_comm():
+    """A comm is a Jupyter object for communicating between a notebook and kernel.
 
-    J_LOGGER.info("IPYTHON: Registering Comms")
+    Set up this object with event handlers."""
+    global merge_complete
+
+    logger.info("IPYTHON: Registering Comms")
 
     comm_target_name = COMM_NAME
 
@@ -168,34 +197,30 @@ def make_comm() -> None:
 
     @jupyter_comm.on_msg
     def _recv(msg):
+        global merge_complete
         if _get_command(msg) == "merge_notebooks":
-            J_LOGGER.info("GOT UPDATE STATUS")
+            logger.info("GOT UPDATE STATUS")
             merge_notebooks(jupyter_comm, msg["content"]["data"])
             return
 
-        J_LOGGER.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
-        J_LOGGER.info(msg)
-        J_LOGGER.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
+        if _get_command(msg) == "merge_complete":
+            logger.info("GOT MERGE COMPLETE")
+            merge_complete = True
+            return
 
-    # store comm for access in this thread later
-    _JupyterComm = jupyter_comm
+        logger.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
+        logger.info(msg)
+        logger.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
 
-    J_LOGGER.info("==> Success")
+    logger.info("==> Success")
 
-    return _JupyterComm
-
-
-def get_comm():
-    # global _JupyterComm
-
-    # assert _JupyterComm, "Uh, how did we get None..."
-    # return _JupyterComm
-    return make_comm()
+    return jupyter_comm
 
 
 def update_cell_contents(comm: Comm, result: Dict[str, Any]) -> None:
-    # J_LOGGER.info(Javascript("Jupyter.notebook.get_cells()"))
+    # logger.info(Javascript("Jupyter.notebook.get_cells()"))
     def _transform_jupytext_cells(jupytext_cells) -> List[Dict[str, Any]]:
+        """TODO: what does this do?"""
         return [
             {"index": i, "output": [], **{k: v for (k, v) in x.items() if k not in {"outputs", "metadata"}}}
             for i, x in enumerate(result["cells"])
@@ -203,10 +228,19 @@ def update_cell_contents(comm: Comm, result: Dict[str, Any]) -> None:
 
     comm.send({"command": "start_sync_notebook", "cells": _transform_jupytext_cells(result["cells"])})
 
+    # Wait for the merge_complete flag to get set in the callback.
+    # This way we don't release the lock before syncing is done.
+    for i in range(500):
+        if merge_complete:
+            return
+        time.sleep(0.01)
+    else:
+        logger.warning("Timed out waiting for syncing to complete.")
     # contents = NotebookContents(cells=result["cells"])
 
 
 def get_output_text(javascript_cell) -> Optional[str]:
+    """Get cell output or return None if no output?"""
     output_tuple = javascript_cell.get("outputs", tuple())
     if not output_tuple:
         return None
@@ -226,7 +260,7 @@ def get_output_text(javascript_cell) -> Optional[str]:
     return None
 
 
-@J_LOGGER.catch(reraise=True)
+@logger.catch(reraise=True)
 def merge_notebooks(comm: Comm, result: Dict[str, Any]) -> None:
     javascript_cells = result["javascript_cells"]
     current_notebook = NotebookContents(
@@ -245,13 +279,16 @@ def merge_notebooks(comm: Comm, result: Dict[str, Any]) -> None:
     new_notebook = NotebookContents(cells=[JupyterCell(**x) for x in result["new_notebook"]])
 
     opcodes = opcode_merge_cell_contents(current_notebook, new_notebook)
-    J_LOGGER.info("Performing Opcodes...")
-    J_LOGGER.info(opcodes)
+    logger.info("Performing Opcodes...")
+    logger.info(opcodes)
 
     net_shift = 0
     for op_action in opcodes:
         net_shift = perform_op_code(comm, op_action, current_notebook, new_notebook, net_shift)
 
+    logger.info("sending finish_merge command")
+    comm.send({"command": "finish_merge"})
+
 
 def perform_op_code(
     comm: Comm,
@@ -272,7 +309,7 @@ def perform_op_code(
         pass
 
     elif op_action.op_code == OpCodes.DELETE:
-        J_LOGGER.info(f"Performing Delete: {op_action}")
+        logger.info(f"Performing Delete: {op_action}")
 
         # Since deletion is a bit goofy for jupyter, so it has to be adjusted by net shift thus far.
         cells_to_delete = [x + net_shift for x in range(*op_action.current)]
@@ -281,7 +318,7 @@ def perform_op_code(
         net_shift = net_shift - len(cells_to_delete)
 
     elif op_action.op_code == OpCodes.INSERT:
-        J_LOGGER.info(f"Performing Insert: {op_action}")
+        logger.info(f"Performing Insert: {op_action}")
 
         cells_to_insert = list(range(*op_action.updated))
         for cell_number in cells_to_insert:
